import os
import sys
import logging

from . import config
from .helpers import runSubprocess

class S3:
    def __init__(self):
        self.bucket = os.path.join(config.s3_bucket, '')
        if not runSubprocess(['s3cmd', 'ls', self.bucket], failOnError=False) and not config.dry_run:
            logging.error('s3cmd is not installed or configured properly')
            sys.exit(1)
        logging.info('Using S3 bucket: %s' % self.bucket)

    def pull_rpms(self):
        command = ['s3cmd', 'sync', os.path.join(self.bucket, ''), os.path.join(config.target_rpm_dir, '')]
        if config.dry_run:
            print(*command)
        else:
            runSubprocess(command)

    def push_rpms(self, delete_removed = False):
        command = ['s3cmd', 'sync', os.path.join(config.target_rpm_dir, ''), self.bucket]
        if delete_removed:
            command.append('--delete-removed')
        if config.dry_run:
            print(*command)
        else:
            runSubprocess(command)



import os

from . import config
from .helpers import parseRecipeHeader


class Defaults:
    def __init__(self, name='o2-dataflow'):
        self.name = name
        filename = 'defaults-%s.sh' % name
        self.path = os.path.join(config.ali_prefix, 'alidist', filename)
        parsed = parseRecipeHeader(self.path)
        self.disable = parsed['disable']



import yaml
import subprocess
import logging
import sys


def parseRecipeHeader(path):
    """ Parses header of alidist recipe"""
    tempString = ''
    with open(path, 'r') as file:
        for line in iter(lambda: file.readline().rstrip(), '---'):
            tempString += line + '\n'
    return yaml.safe_load(tempString)


def parseYamlDict(path):
    """ Parses a yaml list into a dict """
    with open(path, 'r') as file:
        return yaml.safe_load(file)


def runSubprocess(command, **flags):
    options = {
        'failOnError': True,
        'forceTty': False
    }
    for option in options:
        if option in flags:
            options[option] = flags[option]
            del flags[option]
    output = ''
    if options['forceTty']:
        command = ["script -q -e -c '" + ' '.join(command) + "'"]

    try:
        logging.debug('Running as subprocess: %s' % (subprocess.list2cmdline(command)))
        output = subprocess.check_output(command, **flags).strip()
        if type(output) != str:
            output = output.decode('ascii')
    except (subprocess.CalledProcessError, FileNotFoundError) as e:
        output = e.output
        if type(output) != str:
            output = output.decode('ascii')
        if 'File already exists' in output:
            logging.warning('The RPM already exists, continuing...')
        else:
            logging.error(output)
            logging.error(e)
            if options['failOnError']:
                logging.critical('Exiting...')
                sys.exit(1)
            else:
                return False
    logging.debug('Subprocess completed with output length %d' % len(output))
    if (len(output) < 1):
        output = True
    return output



import os
import logging
import sys

from . import config
from .helpers import runSubprocess


class Module:

    def __init__(self, name):
        self.name = name
        os.environ["MODULEPATH"] = os.path.join(config.ali_prefix, 'sw/MODULES', config.architecture)

    def versions(self, filter = True):
        """ List of available versions of given package """
        avail = runSubprocess(['module', '-t', 'avail', self.name, '--no-pager'], forceTty=True, shell=True)
        if not isinstance(avail, str):
            return []
        versions = []
        for entry in (avail.strip().split()[1:]):
            split = entry.rsplit('/')
            # modules wrongly match prefix, eg. Python matches Python-modules, etc..
            if split[0] != self.name and filter:
                continue
            versions.append(split[1])
        return versions

    def deps(self, version):
        """ List of dependencies and versions according to modulefile  """
        filtered = []
        deps = runSubprocess(['module', 'display', 'BASE/1.0', self.name + '/' + version, '--no-pager'], forceTty=True, shell=True).strip()
        for dep in deps.split('\n'):
            if len(dep) > 1 and dep.split()[0] == 'module':
                filtered += dep.split()[2:]
        return filtered

    def deps_as_dict(self, version):
        """ As above but as dictionary """
        dict = {}
        list = self.deps(version)
        for dep in list:
            temp = dep.split('/')
            dict[temp[0]] = temp[1]
        return dict



import os
import re
import logging
import glob
import sys

from . import config
from .helpers import parseYamlDict
from .helpers import runSubprocess

class Fpm:
  """ Class to handle fpm invocations """
  def __init__(self):
    self.buildDir = os.path.join(config.ali_prefix, 'sw', config.architecture)
    self.forceBuild = ['--force'] if config.force_build else []
    self.packagePrefix = "o2-"
    self.targetDir = config.target_rpm_dir
    self.skipDeps = ['openmp', 'VecGeom', 'O2-customization']  # as sometimes there's mismatch between "requires" deps and modulefile

    # Create output dir
    if not os.path.exists(self.targetDir):
        os.makedirs(self.targetDir)

    arch_split = config.architecture.split('_')
    self.architecture = arch_split[1].replace('-', '_')
    self.release_suffix = arch_split[0]

    self.runtimeDepsDict = parseYamlDict(os.path.join(config.work_dir, 'runtime.' + self.release_suffix + '.yaml'))
    self.develDepsDict = parseYamlDict(os.path.join(config.work_dir, 'devel.' + self.release_suffix + '.yaml'))

    if config.tag_version:
        self.release_suffix += '_' + config.release_tag.replace('/', '_')

  def run(self, name, version, deps, devel_deps, extra_deps, RPMS_GENERATED = []):
    """ Prepares arguments and executes fpm """
    fullVersion = version
    version = version.replace('local', '', 1)
    # split version and revision
    version_split = version.split('-')
    # split in case of vX.Y.Z-R
    # don't split if e.g. latest-o2-dataflow
    if len(version_split) == 2:
        version = version_split[0]
        iterationArg = ['--iteration', version_split[1].replace('local', '', 1) + '.' + self.release_suffix]
    else:
        iterationArg = ['--iteration', self.release_suffix]

    rpmType = 'dir'
    # prep dir arguments
    # `=.` is necessary so that the complete source dir path
    # is not replicated and the target dir consists of only
    # prefix + package name + dir
    package_dir = self.buildDir + "/" + name + "/" + fullVersion + "/"
    subdirs = ['bin', 'lib', 'lib64', 'etc', 'include', 'bin-safe', 'libexec', 'WebDID', 'plugins', 'cmake', 'share']
    paths = []
    for subdir in subdirs:
        if os.path.exists(package_dir + subdir):
            if subdir == 'etc' and os.listdir(package_dir + subdir) == ['modulefiles', 'profile.d']:
                continue
            paths.append(package_dir + subdir + '=.')

    # if there are not dirs change RPM type to 'empty'
    if not paths:
        rpmType = 'empty'

    logging.info('Processing %s version %s:\n' \
                 ' - deps: %s\n - extra deps: %s\n - devel deps: %s\n' \
                 % (name, version, list(deps.items()), extra_deps, list({**devel_deps, **self.convertDepsToDevel(deps)}.items())))

    # Handle extra_deps
    extra_deps = [v for dep in extra_deps for v in ('--depends', dep.replace('local', '', 1))]

    packageName = self.packagePrefix + name
    fpmCommand = ['fpm',
      '-s', rpmType,
      '-t', 'rpm',
      '--log', 'warn',
      *self.translateDepsToArgs(deps),
      *extra_deps,
      *self.forceBuild,
      '-p', self.targetDir,
      '--architecture', self.architecture,
      '--prefix', '/opt/o2/',
      '-n', packageName,
      '--version', version,
      *iterationArg,
      '--no-rpm-auto-add-directories',
      '--template-scripts',
      '--after-install', os.path.join(config.work_dir, 'after_install_template.sh'),
      '--after-remove', os.path.join(config.work_dir, 'after_remove_template.sh')]

    if paths:
        fpmCommand.extend(['--exclude=*/modulefiles', '--exclude=*/profile.d', '--exclude=*/info/dir', '--exclude=*/version'])
        fpmCommand.extend(paths)

    develPackageName = packageName + '-devel'
    fpmCommandDevel = ['fpm',
      '-s', 'empty',
      '-t', 'rpm',
      '--log', 'warn',
      '--architecture', self.architecture,
      '--prefix', '/opt/o2/',
      '-n', develPackageName,
      '--version', version,
      '-p', self.targetDir,
      *iterationArg,
      '--template-scripts',
      '--after-install', os.path.join(config.work_dir, 'after_install_template.sh'),
      '--after-remove', os.path.join(config.work_dir, 'after_remove_template.sh'),
      *self.translateDevelDepsToArgs(deps),
      *self.translateDevelDepsToArgs(devel_deps),
      *self.translateDepsToArgs(self.convertDepsToDevel(deps))
    ]

    if not config.dry_run:
      self.generateRpm(packageName, fpmCommand, RPMS_GENERATED)
      if config.devel:
          self.generateRpm(develPackageName, fpmCommandDevel, RPMS_GENERATED)
    else:
      print(*fpmCommand)
      if config.devel:
          print(*fpmCommandDevel)
      return ''

  # Remove system deps and append -devel to names
  def convertDepsToDevel(self, deps):
        return {name+'-devel':version for name, version in deps.items() if version != 'from_system'}

  # Run subprocess to generate RPM
  def generateRpm(self, name, command, RPMS_GENERATED):
      logging.info('Generating RPM: %s', name)
      ret = runSubprocess(command, failOnError=False)
      if not ret:
          logging.warn('Generation of the RPM skipped')
      else:
          # try to parse the generated RPM path from the fpm output
          match = re.search(':path=>\"(.*)\"}', ret)
          generatedRpmPath = match.group(1) if match else ''
          RPMS_GENERATED.append(generatedRpmPath)

  # Convert map name:version to fpm compatible command line params
  def translateDevelDepsToArgs(self, deps):
    depsArg = []
    for name, version in deps.items():
        if name not in self.skipDeps:
            if name in self.develDepsDict:
                for dep in self.develDepsDict.get(name):
                    depsArg.extend(['--depends', dep])
            else:
                depsArg.extend(['--depends', self.packagePrefix + name])
    return depsArg

  # Convert map name:version to fpm compatible command line params
  def translateDepsToArgs(self, deps):
    """ Prepares dependency arguments for fpm """
    depsArg = []
    for depName, depVersion in deps.items():
      if depName not in self.skipDeps:
        if (depVersion != "from_system"):
            countDashes = depVersion.count('-')
            if countDashes > 1:
                depVersion = depVersion.replace('-', '_', countDashes - 1)
            depsArg.extend(['--depends', self.packagePrefix + depName + ' >= ' + depVersion.replace('local', '', 1)])
        else:  # coming from system, get name from dict
          if depName not in self.runtimeDepsDict:
            logging.critical('Could not find system dependency: %s \nDoes it suppose to come from aliBuild?' % (depName))
            sys.exit(1)
          for dep in self.runtimeDepsDict.get(depName):
            depsArg.extend(['--depends', dep])
    return depsArg

  def validateRpms(self):
      """ Validates the generated RPMs by running a dry yum install """

      # get the actual RPM paths to pass to yum and avoid subprocess using shell
      rpmPaths = glob.glob(os.path.join(config.target_rpm_dir, '*.rpm'))
      if rpmPaths == []:
          logging.error("No RPMs found under %s, exiting", config.target_rpm_dir)
          return
      logging.info("Validating the RPMs under %s (will ask for sudo))", config.target_rpm_dir)
      logging.debug("RPMs: \n%s", '\n'.join(rpmPaths))
      yumCommand = ['sudo', 'yum', '-y', 'install',
                    *rpmPaths,
                    '--setopt', 'tsflags=test',
                    '--setopt', 'skip_missing_names_on_install=False']
      if config.dry_run:
          print(*yumCommand)
      else:
          runSubprocess(yumCommand)
